#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <math.h>
#include "Motor.h"

//staticな変数の定義
double Motor::kp=55;
double Motor::ki=5;
int Motor::idCounter=0;
double Motor::tireParam;

Motor::Motor(const char* fd_path, double wheel_radius, double g_ratio, double enc_count){
	this->fd = open(fd_path, O_RDWR);
	this->id=idCounter;
	idCounter++;
	if(this->isFdError()){
		fprintf(stderr,"Motor id:%d fd open error\n",this->id);
	}

	tireParam = 2*M_PI*wheel_radius*g_ratio/enc_count;
}

Motor::~Motor(){
	printf("deleting Motor id:%d ...",this->id);
	if(!this->isFdError()){
		close(this->fd);
		printf("OK.\n");
	}else{
		printf("NG.\n");
		fprintf(stderr,"\tMotor id:%d fd < 0\n",this->id);
	}
}

bool Motor::isFdError(){
	return fd < 0;
}

int Motor::getRotaryEncoder(){
	MotorInfo minfo;
	recvStatusPacket(&(minfo.speed),&(minfo.pwm),&(minfo.puls));
	return minfo.puls*tireParam;
}

int Motor::request (const void *req, void *res, size_t req_len, size_t res_len)
{
	int err;
	uint8_t response;
	errno=0;

	if ( req == NULL ) {
		errno = EINVAL;
		return -1;
	}

	if ( res == NULL ) {
		if ( res_len <= sizeof(response) ) {
			res = &response;
			res_len = sizeof(response);
		}
		else {
			errno = EINVAL;
			return -1;
		}
	}
	assert(res != NULL);

	err = write(this->fd, req, req_len); //リクエストを送信する.
	if ( err < 0 ) {
		err = errno;
		perror("Send");
		fprintf(stderr, "errno : %d\n", err);
		errno = err;
		return -1;
	}

	err = read(this->fd, res, res_len); // 返答を受信する．
	if ( err < 0 ) {
		err = errno;
		perror("Recv");
		errno = err;
		return -1;
	}

	// 返答がACKであるか確認する.
	if ( *(uint8_t *)res != (*(uint8_t *)req | MG3_MOTOR_PKT_ACK) ) {
		errno = EBADRQC; // 不正なリクエストコード
		perror("ReqCode");
		errno = EBADRQC; // 不正なリクエストコード
		return -1;
	}

	return 0;
}

int Motor::sendControlPacket (uint8_t speed)
{
	struct {
		uint8_t command;
		uint8_t spped;
		uint8_t kp;
		uint8_t ki;
	} req = {this->MG3_MOTOR_PKT_SETCTL, speed, this->kp, this->ki};

	assert(sizeof(req) == 4); // パケット長が4であるか確認する.

	return request(&req, NULL, sizeof(req), 0); // モータに制御値を送る．
}

int Motor::recvStatusPacket (int8_t *speed, uint16_t *pwm, int32_t *round)
{
	struct {
		uint8_t response;
		int8_t speed;
		uint16_t pwm;
		int32_t total_round;
	} response;

	assert(sizeof(response) == 8);	// パケット長が8であるか確認する.

	uint8_t req = this->MG3_MOTOR_PKT_GETST;
	int err = request(&req, &response, sizeof(req), sizeof(response)); // モータのステータスを取得する.
	if ( err < 0 ) {
		return err;
	}
	*speed = response.speed;
	*pwm   = response.pwm;
	*round = response.total_round;

	return err;
}

int Motor::sendStartPacket(){
	uint8_t req = this->MG3_MOTOR_PKT_START;
	return request(&req, NULL, 1, 0);
}

int Motor::sendStopPacket(){
	uint8_t req = this->MG3_MOTOR_PKT_STOP;
	return request(&req, NULL, 1, 0);
}

int Motor::sendClearPacket(){
	uint8_t req = this->MG3_MOTOR_PKT_PING;
	return request(&req, NULL, 1, 0);
}
