#ifndef INCLUDED_MOTOR_H
#define INCLUDED_MOTOR_H

#include <stdint.h>
#include <stdlib.h>

class Motor{
public:	
	static const int VLMAX = 80; // = 127-5
	static const int VLMIN = 3;

	//wheel_radius タイヤ径(mm),g_ratio モータとエンコーダの直径比率,enc_count １回転のエンコーダパルス値
	Motor(const char* fd_path,double wheel_radius=57.0,double g_ratio=1.43,double enc_count=360.0);
	~Motor();

	// モータに制御パラメータを送るためのルーチン．
	// オープンされたデバイスファイルthis->fdから,設定したい速度speed，PI制御の各ゲインkp，kiを設定する．返り値については，Motor::request()を参照．
	int sendControlPacket(uint8_t speed);
	// 現在にモータの状態を得るためのルーチン.
	// オープンされたデバイスファイルthis->fdから,回転速度speed，pwmのパルス幅，総回転数roundの値を取得する.返り値については，Motor::request()を参照．
	int recvStatusPacket(int8_t *speed, uint16_t *pwm, int32_t *round);
	// モータに回転を開始させるルーチン.返り値については，Motor::request()を参照．
	int sendStartPacket();
	// モータの回転を停止させるルーチン.返り値については，Motor::request()を参照．
	int sendStopPacket();
	// REカウント値のクリア.返り値については，Motor::request()を参照．
	int sendClearPacket();

	//ロータリーエンコーダの値を取得
	int getRotaryEncoder();
	bool isFdError();
private:
	typedef struct _MotorInfo{
		int8_t speed;
		uint16_t pwm;
		int32_t puls;
	}MotorInfo;

	static double tireParam; //タイヤパラメータ（単位パルスあたりに進む距離）
	static int idCounter;
	int id;
	int fd;
	static double kp;
	static double ki;
	static const int MG3_MOTOR_PKT_PING		= 0x00;
	static const int MG3_MOTOR_PKT_SETCTL	= 0x01;
	static const int MG3_MOTOR_PKT_GETST	= 0x02;
	static const int MG3_MOTOR_PKT_START	= 0x04;
	static const int MG3_MOTOR_PKT_STOP		= 0x08;
	//static const int MG3_MOTOR_PKT_GETCTL	= 0x10;
	static const int MG3_MOTOR_PKT_PERIOD	= 0x20;
	static const int MG3_MOTOR_PKT_ACK		= 0x80;

	// モータと通信をするための基本ルーチン
	// リクエストパケットreqをモータへ送り，返答パケットresを送る.
	// それぞれのパケットの長さは，req_len，res_lenで指定する．
	// 成功すれば，0が返り，失敗すると負数が返る．
	// 失敗したときには，errnoにエラーコードがセットされる．
	// reqパケットの内容が不正な時にはEBADRQCがセットされる.
	// その他の引数が不正な場合，EINVALがセットされる.
	// その他のエラーコードについては，read(2)，write(2)を参照．
	int request (const void *req, void *res, size_t req_len, size_t res_len);
};

/*
 * Protocol
 * # Ping
 * Req: 0x00
 * Res: 0x80
 *
 * # 制御パラメータの設定
 * Req: 0x01, 目標値, Kp, Ki
 * Res: 0x81
 *  目標値: -128?127(パルス/100ms)
 *  Kp: 0?255
 *  Ki: 0?255
 *
 * # モータの状態取得
 * Req: 0x02
 * Res: 0x82, 速度, PWMデューティ(2Byte), 累積カウント(4Byte)
 *  速度: -128?127[パルス/100ms]
 *  PWM duty: 0?255 (duty比: 0?100%)(要検討)
 *  累積カウント: -2^31?2^31-1[パルス]
 *
 * # スタート
 * Req: 0x04
 * Res: 0x84
 *
 * # ストップ
 * Req: 0x08
 * Res: 0x88
 *
 * # 制御パラメータの取得
 * Req: 0x10
 * Res: 0x90, 目標値, Kp, Ki
 *
 * # Unknown
 * Req: 0x??
 * Res: 0x40 | Req
 */

#endif
