#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <pthread.h>
#include <fcntl.h>
#include <string.h>

#include "MotorController.h"

int MotorController::idCounter=0;
/*座標系について
 * 右向きにx軸，上向きにy軸をとった直交座標系を考えるものとする。
 * */
MotorController::MotorController(const char* l_motor_path, const char* r_motor_path, double mirs_radius, double first_pos_x, double first_pos_y, double first_theta){
	this->mirs_radius = mirs_radius;
	strcpy(this->lMotorPath,l_motor_path);
	strcpy(this->rMotorPath,r_motor_path);
	this->lMotorp = new Motor(l_motor_path);
	this->rMotorp = new Motor(r_motor_path);

	this->id = idCounter;
	idCounter++;

	this->startX = this->posX = first_pos_x;
	this->startY = this->posY = first_pos_y;
	this->theta = first_theta;
	this->oldRotaryEncoderValueL = this->rotaryEncoderValueL = this->oldRotaryEncoderValueR = this->rotaryEncoderValueR = 0;
	this->start();
	this->brake();
}

MotorController::~MotorController(){
	printf("deleting MotorController id:%d\n",this->id);
	this->stop();
	delete this->lMotorp;
	delete this->rMotorp;
}

/*データゲット系関数 関数からアクセスするほうがいざというときに便利です*/

double MotorController::getThetaRad(){ return this->theta; }
double MotorController::getThetaDeg(){ return this->theta/M_PI*180.0; }
double MotorController::getPosX(){ return this->posX; }
double MotorController::getPosY(){ return this->posY; }

/*初期化等の関数*/

bool MotorController::isFdError(){
	return this->lMotorp->isFdError() || this->rMotorp->isFdError();
}
void MotorController::initialize(){
	this->lMotorp->sendClearPacket();
	this->rMotorp->sendClearPacket();
	this->rotaryEncoderValueL = this->rotaryEncoderValueR = 0.0;
	this->diffPrev=this->diffSum=0;
	this->startX = this->posX;
	this->startY = this->posY;
}
void MotorController::setStartXY(double start_x,double start_y){
	this->startX = start_x;
	this->startY = start_y;
}

/*走行系関数*/

int MotorController::straight(double speed,double dest, double accomplishment_value){
	return this->std(speed,dest,dest,accomplishment_value);
}
int MotorController::straightKeepDeg(double speed,double dest,double desire_deg,double Ktheta,double Kroute, double accomplishment_value){
	return this->std(speed,dest,dest,accomplishment_value,true,desire_deg,Ktheta,Kroute);
}
int MotorController::straightKeepUss(double speed,double dest, double uss_l_value,double desire_uss_l_value, double Kuss, double accomplishment_value){
	return this->std(speed,dest,dest,accomplishment_value,false,0,0,0,true,desire_uss_l_value,uss_l_value,Kuss);
}
int MotorController::rotate(double speed,double deg, double accomplishment_value){
	return this->std(speed,-deg*M_PI/180.0*this->mirs_radius,deg*M_PI/180.0*this->mirs_radius,accomplishment_value);
}
int MotorController::circle(double speed_lower, double deg, double radius, int rot_direction, double accomplishment_value){
	if( abs(rot_direction) != 1 ){
		fprintf(stderr,"please input 1 or -1\n");
		return 0;
	}
	double dest_l = (deg*M_PI/180) * (radius + this->mirs_radius*sign(rot_direction));
	double dest_r = (deg*M_PI/180) * (radius - this->mirs_radius*sign(rot_direction));
	return this->std(speed_lower,dest_l,dest_r,accomplishment_value);
}
void MotorController::brake(){
	double err1 = this->lMotorp->sendControlPacket(0); //speed値をMCBに送信
	double err2 = this->rMotorp->sendControlPacket(0); //speed値をMCBに送信
	if(err1==-1 || err2==-1){
		ioerrorRecovery();
	}
}

/*以下private*/

void MotorController::start(){ //motorの初期化
	this->lMotorp->sendStartPacket();
	this->rMotorp->sendStartPacket();
}
void MotorController::stop(){ //motorの終了
	this->lMotorp->sendStopPacket();
	this->rMotorp->sendStopPacket();
}

void MotorController::update(){
	this->oldRotaryEncoderValueL = this->rotaryEncoderValueL;
	this->oldRotaryEncoderValueR = this->rotaryEncoderValueR;
	this->rotaryEncoderValueL = this->lMotorp->getRotaryEncoder();
	this->rotaryEncoderValueR = this->rMotorp->getRotaryEncoder();

	double d_rotaryEncoderValueL = this->rotaryEncoderValueL - this->oldRotaryEncoderValueL;
	double d_rotaryEncoderValueR = this->rotaryEncoderValueR - this->oldRotaryEncoderValueR;
	if(fabs(d_rotaryEncoderValueL) > 50 || fabs(d_rotaryEncoderValueR) > 50){
		debug_printf("rotary encoder catch noise.\n");
	}else{
		double d_theta= (d_rotaryEncoderValueR-d_rotaryEncoderValueL)/(2.0*this->mirs_radius);
		this->theta += d_theta;
		debug_printf("d_theta:%.5f\n",d_theta);
		while(this->theta >= 2*M_PI){ this->theta -= 2*M_PI ;}
		while(this->theta <= 0){ this->theta += 2*M_PI ;}
		debug_printf("now deg:%.5f,",this->getThetaDeg());
		this->posX += (d_rotaryEncoderValueL+d_rotaryEncoderValueR)/2.0*cos(theta);
		debug_printf("posX:%.5f,",this->posX);
		this->posY += (d_rotaryEncoderValueL+d_rotaryEncoderValueR)/2.0*sin(theta);
		debug_printf("posY:%.5f\n",this->posY);
	}
}

void MotorController::ioerrorRecovery(){
	fprintf(stderr,"run caught error\n");
	delete this->lMotorp;
	delete this->rMotorp;
	this->lMotorp = new Motor(lMotorPath);
	this->rMotorp = new Motor(rMotorPath);
}

double MotorController::runPidV(double diff, double kp, double kd, double ki){
	diffSum+=diff;
	double ret = kp*diff + kd*(diff-diffPrev) + ki*diffSum;
	diffPrev=diff;
	return ret;
}

char MotorController::sign(double num){
	return (num >= 0)? 1:-1;
}

double MotorController::controlSpeedRange(double speed){
	if(abs(speed) > Motor::VLMAX)      speed = Motor::VLMAX * sign(speed);
	else if(abs(speed) < Motor::VLMIN) speed = Motor::VLMIN * sign(speed);
	return speed;
}

int MotorController::std(double speed_lower, double dest_l, double dest_r, double accomplishment_value, bool keep_deg_flag, double desire_deg, double Ktheta, double Kroute, bool keep_uss_flag, double uss_l_desire_value, double uss_l_value, double Kuss){
	double speed_l,speed_r;
	double ratio     = (dest_r-this->rotaryEncoderValueR)/(dest_l-this->rotaryEncoderValueL);
	speed_lower = controlSpeedRange( speed_lower ); //speed_lowerの大きさを制限する。

	//出力がspeed_lower以下にならないようにしながら，モータ出力の比をratioにする．
	if( ratio*sign(ratio) >= 1 ){ 
		speed_l = speed_lower;
		speed_r = speed_lower*fabs(ratio);
	}else{
		speed_l = speed_lower/fabs(ratio);
		speed_r = speed_lower;
	}
	speed_l *= sign(dest_l-this->rotaryEncoderValueL); //スピードの符号を
	speed_r *= sign(dest_r-this->rotaryEncoderValueR); //目標値に向かうようにとる

	this->update(); //ロータリーエンコーダ値の更新と位置計算

	double speed_l_pid,speed_r_pid;
	speed_l_pid = speed_r_pid = 0;
	if(keep_deg_flag){
		double deg2 = this->getThetaDeg();
		double deg3 = atan2( this->posY-startY, this->posX-startX ) / M_PI*180.0; //(startX,startY)とMIRSの現在の位置を通る直線のx軸に対する角度 -180~180
		//debug_printf("desire_deg:%.5f,deg2:%.5f,deg3:%.5f\n",desire_deg,deg2,deg3);
		double ddeg2 = deg2 - desire_deg;
		double ddeg3 = deg3 - desire_deg;
		while(ddeg2 > 180){  ddeg2-=360 ;}
		while(ddeg3 > 180){  ddeg3-=360 ;}
		while(ddeg2 < -180){ ddeg2+=360 ;}
		while(ddeg3 < -180){ ddeg3+=360 ;}
		//debug_printf("deg2-desire_deg:%.5f,deg3-desire_deg:%.5f\n",ddeg2,ddeg3);
		speed_l_pid += Ktheta*( deg2 - desire_deg ); //deg2-deg==0となるようにPID制御
		speed_l_pid += Kroute*( deg3 - desire_deg ); //deg3-deg==0となるようにPID制御
		speed_r_pid += Ktheta*( desire_deg - deg2 ); //deg2-deg==0となるようにPID制御
		speed_r_pid += Kroute*( desire_deg - deg3 ); //deg3-deg==0となるようにPID制御
	}
	if(keep_uss_flag){
		speed_l_pid += Kuss * ( uss_l_value - uss_l_desire_value );
		speed_r_pid += Kuss * ( uss_l_desire_value - uss_l_value );
	}

	//debug_printf("uss_l_value:%.5f\n",uss_l_value);
	//debug_printf("rotaryEncoderValueL:%f,rotaryEncoderValueR:%f\n",this->rotaryEncoderValueL,this->rotaryEncoderValueR);
	//debug_printf("dest_l:%f,dest_r:%f,speed_lower:%f\n",dest_l,dest_r,speed_lower);
	//debug_printf("speed_l_pid:%f,speed_r_pid:%f\n",speed_l_pid,speed_r_pid);

	char err1 = this->lMotorp->sendControlPacket( speed_l + speed_l_pid ); //speed値をMCBに送信
	char err2 = this->rMotorp->sendControlPacket( speed_r + speed_r_pid ); //speed値をMCBに送信
	if(err1==-1 || err2==-1) ioerrorRecovery();

	/*目標達成判定*/
	if(abs(this->rotaryEncoderValueL-dest_l) < accomplishment_value 
		&& abs(this->rotaryEncoderValueR-dest_r) < accomplishment_value)
	{
		return 1;
	}else{
		return 0;
	}
}

