#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""このモジュールはモータを制御する。"""
import math, os, struct, time

class Motor(object):
    """
    モータを制御するためのクラス。モータ制御ボード(MCB)との通信を行う。
    rotaryencoder を re と省略する
    re_resolvingpower_deg はロータリーエンコーダ固有の「1パルスあたりの角度[deg]」を表す
    実質的な分解能は axis_radius/re_pulley_radius*re_resolvingpower_deg である。
    """
    PKT_CLEAR = 0x00
    PKT_SETCONTROL = 0x01
    PKT_GETSTATUS = 0x02
    PKT_START = 0x04
    PKT_STOP = 0x08
    PKT_CORRECT = 0x80
    #MCB_KP = 50
    #MCB_KI = 17
    MCB_KP = 10
    MCB_KI = 6
    TIRE_RADIUS = 56.8
    AXIS_RADIUS = 5.5
    RE_DEG_PER_SINGLEPULS = 1
    def __init__(self, path, re_pulley_radius=3.85):
        distance_per_deg = 2*math.pi*self.TIRE_RADIUS / 360
        self.re_pulley_radius = re_pulley_radius
        ratio_axis_pulley = self.AXIS_RADIUS / self.re_pulley_radius
        deg_per_single_puls = ratio_axis_pulley * self.RE_DEG_PER_SINGLEPULS
        self.distance_per_single_puls = distance_per_deg * deg_per_single_puls
        self.path = path
        self.file_discriptor = self._open_path(os.O_RDWR)
        self.distance = 0    #[mm]
        self.distance_old = 0 #[mm]
        self.d_distance = 0
        self.oldres = None
        self._start()
        self.clear()
        self.control(0)

    def __del__(self):
        """終了処理"""
        self._stop()
        self._close()
    def update(self):
        """ロータリエンコーダによる距離計測結果を取得、更新する。"""
        #please call this method short term for easily noise delete.
        self.distance_old = self.distance
        self.distance = self._getDistance()
        self.d_distance = self.distance - self.distance_old
        if 100 > abs(self.d_distance) > 50:
            print(" rotary encoder warning (too big balue get).")
        if abs(self.d_distance) > 100: #原因不明の大きな値を得ることがあるので，その値を除外する．解析，除去したい．
            print(" rotary encoder error (too big value get)." +
                   str(self.d_distance/self.distance_per_single_puls))
            self.d_distance = 0
    def clear(self):
        """MCBが保持するロータリエンコーダの総パルス数をクリアする。"""
        self._request(send_data=chr(self.PKT_CLEAR), response_size=1)
        self.distance = 0
    def isComplete(self, dest, complete_delta=5):
        """モータの進んだ距離が目標値(dest)に十分近い(complete_deltaの範囲)ことを判定する"""
        assert complete_delta >= 0, "complete_delta >= 0."
        return abs(dest - self.distance) < complete_delta
    def control(self, speed): #speedは[cm/s]
        """MCBにスピード値を送信する。送信は1byteの整数で行う。"""
        if speed < -128:
            speed = -128
        if speed > 127:
            speed = 127
        snd = struct.pack("BbBB", self.PKT_SETCONTROL, round(speed),
                          self.MCB_KP, self.MCB_KI)
        self._request(send_data=snd, response_size=1)

    def _open_path(self, flag): #virtualmotorでオーバーライドするために定義
        """ファイルオープンの抽象化のための関数"""
        return os.open(self.path, flag)
    def _close(self):
        """ファイルクローズの抽象化を行う"""
        os.close(self.file_discriptor)
    def _recovery(self):
        """MCBあるいは通信用デバイスドライバの復旧を行う。"""
        #self._stop()
        self._close()
        self.file_discriptor = self._open_path(os.O_RDWR)
        self._start()
        #self.control(0) #stopした方がいい？
    def _getDistance(self): #戻り値は[mm]
        """モータが進んだ距離を得る。"""
        res = self._request(send_data=chr(self.PKT_GETSTATUS), response_size=8)
        #print struct.unpack("BbHi", res)[1]
        if res == None or len(res) < 8:
            print "error: response is short"
            self._recovery()
            return 0
        return self.distance_per_single_puls * struct.unpack("BbHi", res)[3]
    def _start(self):
        """MCBに処理の開始を命ずる"""
        self._request(send_data=chr(self.PKT_START), response_size=1)
    def _stop(self):
        """MCBに処理の中断を命ずる"""
        self._request(send_data=chr(self.PKT_STOP), response_size=1)
    def _request(self, send_data, response_size):
        """MCBに命令を送信する"""
        assert response_size > 0, "response_size is wrong"
        try:
            os.write(self.file_discriptor, send_data)
            res = os.read(self.file_discriptor, response_size)
            if ord(res[0]) != (ord(send_data[0]) | self.PKT_CORRECT):
                print "warning: response is wrong"
                return self.oldres
            return res
        except IOError:
            self._recovery()
            return self.oldres
        except OSError:
            self._recovery()
            return self.oldres

def hw_test():
    """ハードウェアのテスト用の関数"""
    lmotor = Motor("/dev/mirS0")
    rmotor = Motor("/dev/mirS1")
    rl = raw_input("rl=? >")
    numstr = raw_input("speed=? >")
    if rl == "r":
        rmotor.control(int(numstr))
    elif rl == "l":
        lmotor.control(int(numstr))
    elif rl == "rl":
        rmotor.control(int(numstr))
        lmotor.control(int(numstr))
    for i in range(1000):
        rmotor.update()
        lmotor.update()

if __name__ == "__main__":
    hw_test()
