#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
モータをセットでコントロールするためのモジュール。
また、モータ出力の結果として得られる位置も扱う。
"""
import virtualmotor, motor, position, math, copy

def _sign(num):
    """符号を得る"""
    return 1 if num >= 0 else -1
def _ratio_abs(numerator, denominator):
    """比を得る"""
    if denominator == 0:
        denominator = 0.000001
    return abs(numerator / denominator)
def _restrict(value, value_max):
    """数値valueの大きさをvalue_maxに制限する"""
    if abs(value) > value_max:
        return value_max * _sign(value)
    else:
        return value
def _pid_controller(desired_value, error, kparam):
    """PID制御の計算を行う"""
    #print error["d"]*kparam["d"], error["p"]*kparam["p"], error["i"]*kparam["i"]
    pid = sum([error[key]*kparam[key] for key in error.keys()])
    value = desired_value + pid
    if value < 0: #実現可能なスピード比はだいたいわかるので、最小値、最大値を作って制限をかけても良いかもしれない。
        print "ratio < 0. pid param may be too big. ratio = " + str(value)
        return 0.01
    #print value
    return value

class Motors(dict):
    """2つのモータをセットで扱うためのクラス"""
    def __init__(self, path_l, path_r, mode="real"):
        if mode == "virtual":
            dict.__init__(self, dict(l=virtualmotor.VirtualMotor(path_l),
                                     r=virtualmotor.VirtualMotor(path_r)))
        elif mode == "real":
            dict.__init__(self, dict(l=motor.Motor(path_l, re_pulley_radius=3.95),
                                     r=motor.Motor(path_r, re_pulley_radius=3.85)))
        else:
            raise ValueError

    def brake(self):
        """全モータのスピードを0にする"""
        self.control(0, 0)
    def clear(self):
        """MCBが保持するロータリエンコーダの総パルス数をクリアする。"""
        for amotor in self.values():
            amotor.clear()
    def control(self, speed_l, speed_r):
        """MCBにスピード値を送信する。送信は1byteの整数で行う。"""
        for amotor, speed in (self["l"], speed_l), (self["r"], speed_r):
            amotor.control(speed)
    def isComplete(self, dest_l, dest_r, complete_delta=5):
        """全モータの進んだ距離がそれぞれ目標値(dest)に十分近い(complete_deltaの範囲)ことを判定する"""
        return all([amotor.isComplete(dest, complete_delta) for amotor, dest
                     in (self["l"], dest_l), (self["r"], dest_r)])
    def update(self):
        """ロータリエンコーダによる距離計測結果を取得、更新する。"""
        for amotor in self.values():
            amotor.update()

class MotorControllerBase(object):
    """走行の制御を行うクラス。直進には角度のD制御(左右のモータスピードの差のP制御)を使用する"""
    MAXSPEED = 100
    TWO_TIRES_DISTANCE = 300
    def __init__(self, path_l, path_r, mode="real"):
        self.motors = Motors(path_l, path_r, mode)
        self.runningmode = "init"
        self.complete_delta = None
        self.speed_low_abs = None
        self.dest = None
        self.kparam = None
        self.mode = mode

    def setStandard(self, speed_low_abs, dest_l, dest_r, complete_delta=20,
               kparam_d=0, kparam_p=0, kparam_i=0):
        """走行の初期化を行う、標準的な(汎用の)関数"""
        #時間があれば引数をdictに変更する
        assert speed_low_abs >= 0, "speed > 0."
        assert (self.runningmode == "brake" or
                 self.runningmode == "init"), "you don't brake"
        self.runningmode = "set"
        self.speed_low_abs = speed_low_abs
        self.complete_delta = complete_delta
        self.dest = dict(l=dest_l, r=dest_r)
        self.kparam = dict(d=kparam_d, p=kparam_p, i=kparam_i)
        self.motors.clear()
    def setStraight(self, speed_abs, dest, complete_delta=20, kparam_d=0):
        """直進走行の初期化を行う"""
        self.setStandard(speed_abs, dest, dest, complete_delta, kparam_d, 0, 0)
    def setRotate(self, speed_abs, deg, complete_delta_deg=10, kparam_d=0):
        """回転の初期化を行う"""
        deg = position.periodicRestrict(deg, -180.0, 180.0)
        dest_r = deg*math.pi/180*self.TWO_TIRES_DISTANCE/2
        dest_l = -dest_r
        self.setStandard(speed_abs, dest_l, dest_r,
                    complete_delta_deg*math.pi/180*self.TWO_TIRES_DISTANCE/2,
                    kparam_d, 0, 0)
    def run(self):
        """走行のための情報を更新し、スピード値をMCBに送信する"""
        assert (self.runningmode == "set" or
                 self.runningmode == "running"), "you forget to set"
        self.runningmode = "running"
        self.motors.update()
        speed_l, speed_r = self._calcSpeed(self.speed_low_abs,
                                           self.dest["l"], self.dest["r"],
                                           self.kparam)
        #print speed_l, speed_r
        self.motors.control(speed_l, speed_r)
    def isComplete(self):
        """走行の完了を判定する"""
        return self.motors.isComplete(self.dest["l"], self.dest["r"],
                                       self.complete_delta)
    def brake(self):
        """停止する"""
        assert (self.runningmode == "running" or
                 self.runningmode == "set"), "you forget to run"
        self.runningmode = "brake"
        self.motors.brake()

    def _calcSpeedRatioAbs(self, dest_l, dest_r, kparam):
        """左右のスピード比を計算する。絶対値。"""
        ratio = _ratio_abs(dest_r, dest_l)
        error = dict(d=ratio * abs(self.motors["l"].distance) -
                       abs(self.motors["r"].distance))
        return _pid_controller(
            desired_value=ratio,
            error=error,
            kparam=kparam)
    def _calcSpeed(self, speed_low_abs, dest_l, dest_r, kparam):
        """左右のスピードの計算を行い、タプルとして返す"""
        speed_low_abs = _restrict(speed_low_abs, self.MAXSPEED)
        direction_l = _sign(dest_l - self.motors["l"].distance)
        direction_r = _sign(dest_r - self.motors["r"].distance)
        ratio_abs = self._calcSpeedRatioAbs(dest_l, dest_r, kparam)
        #以下 ratio_abs == abs(speed_r/speed_l) を用いて計算
        if ratio_abs >= 1:
            return (direction_l * speed_low_abs,
                     direction_r * speed_low_abs * ratio_abs)
        else:
            return (direction_l * speed_low_abs / ratio_abs,
                     direction_r * speed_low_abs)

class MotorController(MotorControllerBase):
    """走行の制御に加えて、位置の推定も行う。また、直進には角度のPID制御を行う。"""
    def __init__(self, path_l, path_r, first_pos_x=0, first_pos_y=0,
                 first_theta_rad=math.pi/2, mode="real"):
        MotorControllerBase.__init__(self, path_l, path_r, mode)
        self.startPos = position.Position(first_pos_x, first_pos_y,
                                          first_theta_rad)
        self.pos = position.Position(first_pos_x, first_pos_y,
                                     first_theta_rad)

    def run(self):
        """走行のための情報を更新し、スピード値をMCBに送信する"""
        self.update()
        if self.mode == "virtual":
            print self
        MotorControllerBase.run(self)
    def update(self):
        """位置情報の更新を行う"""
        lmotor = self.motors["l"]
        rmotor = self.motors["r"]
        ave_d_distance = (lmotor.d_distance + rmotor.d_distance) / 2
        self.pos.x += ave_d_distance * math.cos(self.pos.theta)
        self.pos.y += ave_d_distance * math.sin(self.pos.theta)
        diff_d_distance = rmotor.d_distance - lmotor.d_distance
        self.pos.theta += diff_d_distance / self.TWO_TIRES_DISTANCE
        self.pos.theta = position.periodicRestrict(self.pos.theta,
                                                   0, 2*math.pi)
    def _calcSpeedRatioAbs(self, dest_l, dest_r, kparam):
        ratio = _ratio_abs(dest_r, dest_l)
        starttheta = self.startPos.getThetaRad()
        theta1 = starttheta - self.pos.getThetaRad()
        theta2 = starttheta - math.atan2(self.pos.y - self.startPos.y,
                                                  self.pos.x - self.startPos.x)
        ave_distance = (self.motors["l"].distance + self.motors["r"].distance) / 2
        error = dict(d=ratio*abs(self.motors["l"].distance) - abs(self.motors["r"].distance),
                     p=position.periodicRestrict(theta1, -math.pi, math.pi),
                     i=ave_distance*math.sin(theta2))
        return _pid_controller(
            desired_value=ratio,
            error=error,
            kparam=kparam)
    def setStartPos(self, pos):
        """直進走行の開始位置と開始角度を指定し、直進を行うように初期化を行う"""
        self.startPos = copy.deepcopy(pos)
    def setRotateAbsolute(self, speed_abs, deg_absolute, complete_delta_deg=10,
                          kparam_d=0):
        """絶対座標の角度指定で回転を行うように初期化を行う"""
        deg = deg_absolute - self.pos.getThetaDeg()
        self.setRotate(speed_abs, deg, complete_delta_deg, kparam_d)
    def setStraightPID(self, speed_abs, dest, startpos, complete_delta=20,
                    kparam_d=0, kparam_p=0, kparam_i=0):
        """
        直進走行の初期化を行う
        バックはdestを負にすれば可能だが、P制御とI制御の動作は保証できない
        """
        self.setStartPos(startpos)
        self.setStandard(speed_abs, dest, dest, complete_delta,
                         kparam_d, kparam_p, kparam_i)
    def __str__(self):
        return str(self.pos)

def simple_running_test():
    """単純な走行テスト"""
    mc = MotorController("/dev/mirS0", "/dev/mirS1", mode="real")

    nextpos = position.Position(0, 500)
    dpos = nextpos - mc.pos
    
    mc.setStraightPID(
        speed_abs=10, dest=abs(dpos),
        startpos=position.Position(0, 0, dpos.getArgRad()),
        kparam_d=0.0, kparam_p=0.6, kparam_i=0.01, complete_delta=30) # slow speed

    #mc.setStraightPID(
    #    speed_abs=40, dest=abs(dpos),
    #    startpos=position.Position(0, 0, dpos.getArgRad()),
    #    kparam_d=0.005, kparam_p=0.8, kparam_i=0.01, complete_delta=40) # slow speed
    while not mc.isComplete():
        mc.run()
        print mc
    
    """while True:
        mc.setRotate(speed_abs=3, deg=90.0, kparam_d=0.001, complete_delta_deg=1)
        while not mc.isComplete():
            mc.run()
            print mc
        mc.brake()
        if raw_input() == "quit": break
    """

if __name__ == "__main__":
    simple_running_test()
