#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""数字認識を行うためのモジュール"""
#import cv
from cv2 import cv

def _scanLineWidth(src1c, range_list, height):
    """横向きに走査する"""
    for i in range_list:
        if src1c[height, i] == 255:
            return 1
    return 0
def _scanLineHeight(src1c, range_list, width):
    """縦向きに走査する"""
    for i in range_list:
        if src1c[i, width] == 255:
            return 1
        return 0
def _numberFlagMatching(flags):
    """画像の捜査結果の情報を引数に取り、数字を返す"""
    if flags == [1, 1, 1, 1, 1, 1, 0, 0]: return 0
    if flags == [1, 1, 1, 1, 1, 1, 1, 1]: return 1
    if flags == [1, 1, 0, 1, 1, 0, 1, 0]: return 2
    if flags == [0, 1, 1, 1, 0, 1, 1, 0]: return 3
    if flags == [0, 1, 1, 0, 0, 1, 1, 0]: return 4
    if flags == [1, 0, 1, 1, 0, 1, 1, 0]: return 5
    if flags == [1, 0, 1, 1, 1, 1, 1, 0]: return 6
    if flags == [1, 1, 1, 0, 0, 0, 0, 0]: return 7
    if flags == [1, 1, 1, 1, 1, 1, 1, 0]: return 8
    if flags == [1, 1, 1, 1, 0, 1, 1, 0]: return 9
    return -1
def _linearScan8(src1c, rect):
    """画像を8カ所走査する"""
    width = rect[2]
    height = rect[3]
    flags = [0]*8
    flags[0] = _scanLineHeight(src1c, range(height/4), width/2)
    flags[1] = _scanLineWidth(src1c, range(width/2, width-1), height/4)
    flags[2] = _scanLineWidth(src1c, range(width/2, width-1), height*3/4)
    flags[3] = _scanLineHeight(src1c, range(height*3/4, height-1), width/2)
    flags[4] = _scanLineWidth(src1c, range(width/2), height*3/4)
    flags[5] = _scanLineWidth(src1c, range(width/2), height/4)
    flags[6] = _scanLineHeight(src1c, range(height/4, height*3/4), width/2)
    if (src1c[height*3/4, width/2] == 255 and
        src1c[height/4, width/2] == 255):
        flags[7] = 1
    return flags

class NumberDetector(object):
    """数字認識を行うためのクラス"""
    def __init__(self, height, width, identification, debug=True):
        self._tmp3c = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 3)
        self._tmp1c = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 1)
        self._output1c = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 1)
        self._binary1c = cv.CreateImage((width, height), cv.IPL_DEPTH_8U, 1)
        self.data = cv.CreateMemStorage()
        self.numbers = []
        self.identification = identification
        self.debug = debug
        self.numbers_count = 2

    def update(self, input3c):
        """画像を処理しやすい形式にして、数字認識を行う"""
        self._binarization(input3c, self._output1c)
        labelling_rects = self._labelling(self._output1c)
        self.numbers = self._numberDetect(self._output1c, labelling_rects, numbers_count=self.numbers_count)
    def clear(self):
        """数字認識の結果をクリアする"""
        self.numbers = []

    def _numberDetect(self, src1c, labelling_rects, numbers_count=2):
        """数字認識を行う"""
        numbers = []
        count = 0
        for rect in labelling_rects:
            cv.SetImageROI(src1c, rect)
            cv.Erode(src1c, src1c, iterations=3)
            self._output(src1c, "trimming" + str(count))
            flags = _linearScan8(src1c, rect)
            number = _numberFlagMatching(flags)
            if number != -1:
                numbers.append(number)
            cv.ResetImageROI(src1c)
            if len(numbers) == numbers_count:
                return numbers
            count += 1
    def _binarization(self, src3c, dst1c):
        """2値化を行う"""
        cv.CvtColor(src3c, dst1c, cv.CV_BGR2GRAY)
        cv.Smooth(dst1c, self._tmp1c, cv.CV_GAUSSIAN, 5, 0, 0, 0)
        cv.Threshold(self._tmp1c, dst1c, 90, 255,
                     cv.CV_THRESH_BINARY_INV | cv.CV_THRESH_OTSU)
    def _labelling(self, src1c, width_minmax=(10, 100),
                   height_minmax=(40, 100)):
        """ラベリングを行う"""
        cv.Copy(self._output1c, self._binary1c)
        if self.debug == True:
            cv.MixChannels([self._output1c], [self._tmp3c], [(0, 2)])
        contours = cv.FindContours(src1c, self.data)
        ret = []
        while contours != None:
            rect = cv.BoundingRect(contours, 1) #int型が返るので、演算には注意
            if (width_minmax[0] < rect[2] < width_minmax[1] and
                height_minmax[0] < rect[3] < height_minmax[1]):
                if self.debug == True:
                    cv.Rectangle(self._tmp3c, (rect[0], rect[1]),
                                 (rect[0] + rect[2], rect[1] + rect[3]),
                                 cv.CV_RGB(255, 255, 255), 1, 8, 0)
                ret.append((rect[0], rect[1], rect[2], rect[3]))
            contours = contours.h_next()
        self._output(self._tmp3c, "labelling")
        cv.Copy(self._binary1c, self._output1c)
        return ret
    def _output(self, out, identification_str):
        """画像として出力する"""
        if self.debug == True:
            cv.SaveImage("data/outputnum" + str(self.identification) +
                         "("+ identification_str +").png", out)
